"""Class for representing models.
"""

from astrocats.catalog.catdict import CatDict, CatDictError
from astrocats.catalog.error import Error
from astrocats.catalog.key import KEY_TYPES, Key, KeyCollection
from astrocats.catalog.realization import Realization


class MODEL(KeyCollection):
    # Strings
    SOURCE = Key('source', KEY_TYPES.STRING, compare=False)
    CODE = Key('code', KEY_TYPES.STRING)
    NAME = Key('name', KEY_TYPES.STRING)
    VERSION = Key('version', KEY_TYPES.STRING)
    DATE = Key('date', KEY_TYPES.STRING)
    DESC = Key('description', KEY_TYPES.STRING, compare=False)
    # Numbers
    ALIAS = Key('alias', KEY_TYPES.NUMERIC, compare=False)
    # Arrays
    REALIZATIONS = Key('realizations', compare=False)
    SETUP = Key('setup', compare=False)


class Model(CatDict):
    """Container for a model with associated metadata.

    `Source` citation required.
    """

    _ALLOW_UNKNOWN_KEYS = True
    _KEYS = MODEL

    def __init__(self, parent, **kwargs):
        self._REQ_KEY_SETS = [[MODEL.SOURCE], [MODEL.ALIAS],
                              [MODEL.NAME, MODEL.CODE]]
        # Note: `_check()` is called at end of `super().__init__`
        super().__init__(parent, **kwargs)
        self.catalog = parent.catalog

        return

    def _check(self):
        """

        """
        # Run the super method
        super()._check()

        err_str = None

        # Future checks go here

        if err_str is not None:
            raise ValueError(err_str)

        return

    def _clean_value_for_key(self, key, value):
        value = super()._clean_value_for_key(key, value)

        return value

    def _init_cat_dict(self, cat_dict_class, key_in_self, **kwargs):
        """Initialize a CatDict object, checking for errors.
        """
        # Catch errors associated with crappy, but not unexpected data
        try:
            new_entry = cat_dict_class(self, key=key_in_self, **kwargs)
        except CatDictError as err:
            if err.warn:
                self._log.info("'{}' Not adding '{}': '{}'".format(self[
                    self._KEYS.NAME], key_in_self, str(err)))
            return None
        return new_entry

    def _add_cat_dict(self,
                      cat_dict_class,
                      key_in_self,
                      check_for_dupes=True,
                      **kwargs):
        """Add a CatDict to this Entry if initialization succeeds and it
        doesn't already exist within the Entry.
        """
        # Try to create a new instance of this subclass of `CatDict`
        new_entry = self._init_cat_dict(cat_dict_class, key_in_self, **kwargs)
        if new_entry is None:
            return False

        # Compare this new entry with all previous entries to make sure is new
        if cat_dict_class != Error:
            for item in self.get(key_in_self, []):
                if new_entry.is_duplicate_of(item):
                    item.append_sources_from(new_entry)
                    # Return the entry in case we want to use any additional
                    # tags to augment the old entry
                    return new_entry

        self.setdefault(key_in_self, []).append(new_entry)

        return True

    def add_realization(self, **kwargs):
        self._add_cat_dict(Realization, self._KEYS.REALIZATIONS, **kwargs)

    def sort_func(self, key):
        if key == self._KEYS.SOURCE:
            return 'zzy'
        if key == self._KEYS.ALIAS:
            return 'zzz'
        return key



"""Class for representing model realizations.
"""

from astrocats.catalog.catdict import CatDict
from astrocats.catalog.key import KEY_TYPES, Key, KeyCollection


class REALIZATION(KeyCollection):
    SCORE = Key('score', KEY_TYPES.NUMERIC)
    PARAMETERS = Key('parameters', KEY_TYPES.DICT)


class Realization(CatDict):
    """Container for a single model realization.
    """

    _ALLOW_UNKNOWN_KEYS = True
    _KEYS = REALIZATION

    def __init__(self, parent, **kwargs):
        self._REQ_KEY_SETS = []
        # Note: `_check()` is called at end of `super().__init__`
        super().__init__(parent, **kwargs)

        return

    def _check(self):
        """

        """
        # Run the super method
        super()._check()

        err_str = None

        if err_str is not None:
            raise ValueError(err_str)

        return

    def _clean_value_for_key(self, key, value):
        value = super()._clean_value_for_key(key, value)

        return value

    def sort_func(self, key):
        return key



"""Entry point for the AstroCats template Catalog.
"""


def main(args, clargs, log):
    from .catalog import Catalog
    from .argshandler import ArgsHandler

    # Create an `ArgsHandler` instance with the appropriate argparse machinery
    args_handler = ArgsHandler(log)
    # Parse the arguments to get the configuration settings
    args = args_handler.load_args(args=args, clargs=clargs)
    # Returns 'None' if no subcommand is given
    if args is None:
        return

    # Create the appropriate type of catalog
    catalog = Catalog(args, log)

    # Run the subcommand given in `args`
    args_handler.run_subcommand(args, catalog)

    return



"""Astrocats: Scripts for creating and analyzing catalogs of astronomical data.
"""

import os

__version__ = '0.2.15'
__author__ = 'James Guillochon'
__license__ = 'MIT'

# Set the path for the user's configuration file
_CONFIG_PATH = os.path.join(os.path.expanduser('~'),
                            '.config', 'astrocats', 'astrocatsrc')



"""WARNING : THIS SCRIPT IS NOT CURRENTLY OPERATIONAL.
"""

import logging
import os

from setuptools import find_packages, setup
from setuptools.command.develop import develop
from setuptools.command.install import install

with open('requirements.txt') as f:
    required = f.read().splitlines()

dir_path = os.path.dirname(os.path.realpath(__file__))
exec(open(os.path.join(dir_path, 'astrocats', '__init__.py')).read())


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def setup_uc():
    from astrocats.main import setup_user_config

    setup_user_config(logging.getLogger())


class PostDevelopCommand(develop):
    """Post-develop command."""

    def run(self):
        setup_uc()
        develop.run(self)


class PostInstallCommand(install):
    """Post-installation command."""

    def run(self):
        setup_uc()
        install.run(self)


setup(
    name="astrocats",
    packages=find_packages(exclude=('*supernovae*', '*tidaldisruptions*',
                                    '*novae*')),
    include_package_data=True,
    version=__version__,  # noqa
    description=("Package for downloading, analyzing, and constructing open "
                 "astronomy catalogs."),
    license=__license__,  # noqa
    author=__author__,  # noqa
    author_email="guillochon@gmail.com",
    install_requires=required,
    url="https://github.com/astrocatalogs/astrocats",
    download_url=(
        'https://github.com/astrocatalogs/astrocats/tarball/' +
        __version__  # noqa
    ),
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
    },
    keywords="astronomy",
    long_description=read('README.md'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5"
    ],
    zip_safe=True)



"""Logging submodule and related functions.
"""

import inspect
import logging
from logging import DEBUG, INFO, WARNING

import numpy as np

_FILE_LEVEL_DEF = DEBUG
_STREAM_LEVEL_DEF = WARNING
_LOADED_LEVEL = INFO


__all__ = ["get_logger", "log_raise", "DEBUG", "WARNING", "INFO"]


class IndentFormatter(logging.Formatter):
    """Logging formatter where the depth of the stack sets the message
    indentation level.
    """

    def __init__(self, fmt=None, datefmt=None):
        logging.Formatter.__init__(self, fmt, datefmt)
        self.baseline = None

    def format(self, rec):
        stack = inspect.stack()
        if self.baseline is None:
            self.baseline = len(stack)
        indent = (len(stack) - self.baseline)
        addSpace = ((indent > 0) & (not rec.msg.startswith(" -")))
        rec.indent = ' -' * indent + ' ' * addSpace
        out = logging.Formatter.format(self, rec)
        del rec.indent
        return out


def get_logger(name=None, stream_fmt=None, file_fmt=None, date_fmt=None,
               stream_level=None, file_level=None,
               tofile=None, tostr=True):
    """Create a standard logger object which logs to file and or stdout stream.

    If a logger has already been created in this session, it is returned
    (unless `name` is given).

    Arguments
    ---------
    name : str,
        Handle for this logger, must be distinct for a distinct logger.
    stream_fmt : str or `None`,
        Format of log messages to stream (stdout).  If `None`, default settings
        are used.
    file_fmt : str or `None`,
        Format of log messages to file.  If `None`, default settings are used.
    date_fmt : str or `None`
        Format of time stamps to stream and/or file.  If `None`, default
        settings are used.
    stream_level : int,
        Logging level for stream.
    file_level : int,
        Logging level for file.
    tofile : str or `None`,
        Filename to log to (turned off if `None`).
    tostr : bool,
        Log to stdout stream.

    Returns
    -------
    logger : ``logging.Logger`` object,
        Logger object to use for logging.

    """
    if tofile is None and not tostr:
        raise ValueError(
            "Must log to something: `tofile` or `tostr` must be `True`.")

    logger = logging.getLogger(name)
    # Add a custom attribute to this `logger` so that we know when an existing
    # one is being returned
    if hasattr(logger, '_OSC_LOGGER'):
        return logger
    else:
        logger._OSC_LOGGER = True

    # Set other custom parameters
    logger._LOADED = _LOADED_LEVEL

    # Make sure handlers don't get duplicated (ipython issue)
    while len(logger.handlers) > 0:
        logger.handlers.pop()
    # Prevents duplication or something something...
    logger.propagate = 0

    # Determine and Set Logging Levels
    if file_level is None:
        file_level = _FILE_LEVEL_DEF
    if stream_level is None:
        stream_level = _STREAM_LEVEL_DEF
    # Logger object must be at minimum level
    logger.setLevel(int(np.min([file_level, stream_level])))

    if date_fmt is None:
        date_fmt = '%Y/%m/%d %H:%M:%S'

    # Log to file
    # -----------
    if tofile is not None:
        if file_fmt is None:
            file_fmt = "%(asctime)s %(levelname)8.8s [%(filename)20.20s:"
            file_fmt += "%(funcName)-20.20s]%(indent)s%(message)s"

        fileFormatter = IndentFormatter(file_fmt, datefmt=date_fmt)
        fileHandler = logging.FileHandler(tofile, 'w')
        fileHandler.setFormatter(fileFormatter)
        fileHandler.setLevel(file_level)
        logger.addHandler(fileHandler)
        #     Store output filename to `logger` object
        logger.filename = tofile

    # Log To stdout
    # -------------
    if tostr:
        if stream_fmt is None:
            stream_fmt = "%(indent)s%(message)s"

        strFormatter = IndentFormatter(stream_fmt, datefmt=date_fmt)
        strHandler = logging.StreamHandler()
        strHandler.setFormatter(strFormatter)
        strHandler.setLevel(stream_level)
        logger.addHandler(strHandler)

    return logger


def log_raise(log, err_str, err_type=RuntimeError):
    """Log an error message and raise an error.

    Arguments
    ---------
    log : `logging.Logger` object
    err_str : str
        Error message to be logged and raised.
    err_type : `Exception` object
        Type of error to raise.

    """
    log.error(err_str)
    # Make sure output is flushed
    # (happens automatically to `StreamHandlers`, but not `FileHandlers`)
    for handle in log.handlers:
        handle.flush()
    # Raise given error
    raise err_type(err_str)
